// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package testdata

// B is declared in a separate file to test that object resolution spans the
// entire package.
var B int
