// +build ignore
// The previous line (and this one up to the warning below) is removed by the
// template generator.

// Code generated from params.templ.go. DO NOT EDIT.

package internal

const (
	Name          = "{{.Name}}"
	K             = {{.K}}
	L             = {{.L}}
	Eta           = {{.Eta}}
	DoubleEtaBits = {{.DoubleEtaBits}}
	Omega         = {{.Omega}}
	Tau           = {{.Tau}}
	Gamma1Bits    = {{.Gamma1Bits}}
	Gamma2        = {{.Gamma2}}
	NIST          = {{.NIST}}
	TRSize        = {{.TRSize}}
	CTildeSize    = {{.CTildeSize}}
)
