// Package fourq provides elliptic curve operations over FourQ curve.
//
// FourQ is a high-speed elliptic curve at the 128-bit security level. This package
// contains an AMD64-optimized implementation. In particular, this package does
// not implement FourQ's endomorphisms or lattice reduction techniques.
//
// References:
//   - https://eprint.iacr.org/2015/565
package fourq
