// Package frodo provides the key encapsulation mechanism FrodoKEM.
//
// Compatible with the implementation submitted to round 3 of the
// NIST PQC competition [1]. This implementation draws heavily from the PQClean
// implementation [2].
//
// References:
//
//	[1] https://frodokem.org/files/FrodoKEM-specification-20210604.pdf
//	[2] https://github.com/PQClean/PQClean/tree/master/crypto_kem/frodokem640shake/clean
package frodo
