// Package vdaf provides verifiable distributed aggregation functions.
//
// This package supports Prio3 as specified in [v13].
// [v13]: https://datatracker.ietf.org/doc/html/draft-irtf-cfrg-vdaf-13
package vdaf
