// Package curve4q implements Diffie-Hellman operations using the FourQ curve
// at the 128-bit security level.
//
// References:
//   - https://eprint.iacr.org/2015/565
//   - https://tools.ietf.org/html/draft-ladd-cfrg-4q-01
package curve4q
