// Package mlkem implements  IND-CCA2 secure ML-KEM key encapsulation
// mechanism (KEM) as defined in FIPS 203.
//
// https://nvlpubs.nist.gov/nistpubs/fips/nist.fips.203.pdf
package mlkem

// See ../kyber/gen.go and ../kyber/kat_test.go.
